# -*- encoding: utf-8 -*-
"""
License: MIT
Copyright (c) 2019 - present AppSeed.us
Copyright (c) 2020 - NeuroAssassin
"""
# Flask related libraries
from flask import Flask, url_for, session, render_template
from flask_babel import Babel
from waitress import serve

# Terminal related libraries
from rich import console, traceback as rtb, progress, logging as richlogging
import logging

# Secret key libraries
from cryptography import fernet
import base64

# Background thread libraries
import threading
import time
import websocket
import traceback

# Base libraries
from babel import Locale
import json
import sys
import os

# Relative imports
from reddash.app.constants import Lock, DEFAULTS, WS_URL, ALLOWED_LOCALES
from reddash.app.tasks import TaskManager
from reddash.app.utils import (
    register_blueprints,
    apply_themes,
    add_constants,
    initialize_babel,
    startup_message,
)


# Logging and terminal set up

log = logging.getLogger("werkzeug")
dashlog = logging.getLogger("reddash")
queuelog = logging.getLogger("waitress.queue")

console = console.Console()
oldexcepthook = rtb.install()
progress_bar = progress.Progress(
    "{task.description}", progress.TextColumn("{task.fields[status]}\n")
)

logging.basicConfig(format="%(message)s", handlers=[richlogging.RichHandler(console=progress_bar)])
queuelog.setLevel(logging.ERROR)

# Base variable setup
app = Flask("reddash", static_folder="app/base/static")
lock = Lock()
babel = Babel()


def create_app(host, port, rpcport, interval, debug, dev):
    url = f"{WS_URL}{rpcport}"

    # Session encoding
    fernet_key = fernet.Fernet.generate_key()
    secret_key = base64.urlsafe_b64decode(fernet_key)

    # JWT encoding
    jwt_fernet_key = fernet.Fernet.generate_key()
    jwt_secret_key = base64.urlsafe_b64decode(jwt_fernet_key)

    # Initialize websocket variables
    app.ws_url = url
    app.ws = None
    app.lock = lock
    app.rpcport = str(rpcport)
    app.rpcversion = 0
    app.interval = interval

    # Initialize core variables
    app.task_manager = TaskManager(app, console, progress_bar)
    app.dashlog = dashlog
    app.progress = progress_bar
    app.running = True
    app.variables = {}
    app.commanddata = {}
    app.cooldowns = {
        "serverprefix": {},
        "adminroles": {},
        "modroles": {},
        "fetchrules": {},
        "fetchtargets": {},
        "fetchcogcommands": {},
        "addrule": {},
        "adddefaultrule": {},
        "removerule": {},
        "removedefaultrule": {},
        "fetchaliases": {},
    }
    app.blacklisted = []

    # Initialize config
    app.config.from_object(__name__)
    app.config["TEMPLATES_AUTO_RELOAD"] = True
    app.config["BABEL_TRANSLATION_DIRECTORIES"] = "app/translations"

    locale_dict = {}
    for locale in ALLOWED_LOCALES:
        l = Locale.parse(locale)
        lang = l.get_language_name(locale)
        if territory := l.get_territory_name(locale):
            lang = f"{l.get_language_name(locale)} - {l.get_territory_name(locale)}"
        locale_dict[locale] = lang

    app.config["LANGUAGES"] = ALLOWED_LOCALES
    app.config["LOCALE_DICT"] = locale_dict
    app.secret_key = secret_key
    app.jwt_secret_key = jwt_secret_key

    babel = Babel(app)

    # Initialize core app functions
    register_blueprints(app)
    apply_themes(app)
    add_constants(app)
    initialize_babel(app, babel)

    if debug:
        log.setLevel(logging.DEBUG)
        dashlog.setLevel(logging.DEBUG)

    kwargs = {"host": host, "port": port, "dev": dev, "debug": debug}
    progress_tasks = startup_message(app, progress_bar, kwargs)

    app.task_manager.start_tasks(progress_tasks)

    if dev:
        app.run(host=host, port=port, debug=debug)
    else:
        serve(app, host=host, port=port, _quiet=True)

    app.dashlog.fatal("Shutting down...")
    app.running = False
    app.task_manager.stop_tasks()

    app.dashlog.fatal("Webserver terminated")



from reddash.app import app
from reddash.app.api import blueprint
from flask import render_template, redirect, url_for, session, request, jsonify, Response, g
from flask_babel import _, refresh
from jinja2 import TemplateNotFound
import traceback
import websocket
import json
import time
import random
import logging
import datetime

dashlog = logging.getLogger("reddash")


def get_result(app, requeststr):
    app.ws.send(json.dumps(requeststr))
    result = json.loads(app.ws.recv())
    if "error" in result:
        if result["error"]["message"] == "Method not found":
            return jsonify({"status": 0, "message": _("Not connected to bot")})
        dashlog.error(result["error"])
        return jsonify({"status": 0, "message": _("Something went wrong")})
    if isinstance(result["result"], dict) and result["result"].get("disconnected", False):
        return jsonify({"status": 0, "message": _("Not connected to bot")})
    return jsonify({"status": 1, "data": result["result"]})


# --------------------------------------- API ---------------------------------------


@blueprint.route("/api/getservers")
def getservers():
    if not session.get("id"):
        return jsonify({"status": 0, "message": _("Not logged in")})
    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC__GET_USERS_SERVERS",
            "params": [str(g.id)],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": _("Not connected to bot")})


@blueprint.route("/api/<guild>/serverprefix", methods=["POST"])
def serverprefix(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": _("Not logged in")})

    if (
        end := app.cooldowns["serverprefix"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["serverprefix"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=5)

    data = request.json
    userid = g.id
    method = "set"
    prefixes = data.get("prefixes")

    if not prefixes:
        return jsonify({"status": 0, "message": _("Prefixes must be specified")})

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_BOTSETTINGS__SERVERPREFIX",
            "params": [str(guild), str(userid), method, prefixes],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/adminroles", methods=["POST"])
def adminroles(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["adminroles"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["adminroles"][session.get("id")] = datetime.datetime.now() + datetime.timedelta(
        seconds=5
    )

    data = request.json
    userid = g.id
    method = "set"
    roles = data.get("roles")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_BOTSETTINGS__ADMINROLES",
            "params": [str(guild), str(userid), method, roles],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/modroles", methods=["POST"])
def modroles(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["modroles"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["modroles"][session.get("id")] = datetime.datetime.now() + datetime.timedelta(
        seconds=5
    )

    data = request.json
    userid = g.id
    method = "set"
    roles = data.get("roles")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_BOTSETTINGS__MODROLES",
            "params": [str(guild), str(userid), method, roles],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/fetchrules", methods=["GET"])
def fetchrules(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["fetchrules"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["fetchrules"][session.get("id")] = datetime.datetime.now() + datetime.timedelta(
        seconds=5
    )

    userid = g.id

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__FETCH_GUILD_RULES",
            "params": [str(guild), str(userid)],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/fetchtargets", methods=["GET"])
def fetchtargets(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["fetchtargets"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=10)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["fetchtargets"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=10)

    userid = g.id

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__FETCH_GUILD_TARGETS",
            "params": [str(guild), str(userid)],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/fetchcogcommands", methods=["GET"])
def fetchcogcommands(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["fetchcogcommands"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["fetchcogcommands"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=5)

    userid = g.id

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__FETCH_COG_COMMANDS",
            "params": [str(guild), str(userid)],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/addrule", methods=["POST"])
def addrule(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["addrule"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["addrule"][session.get("id")] = datetime.datetime.now() + datetime.timedelta(
        seconds=5
    )

    data = request.json

    userid = g.id
    allow_or_deny = data.get("ad")
    who_or_what = data.get("ww")
    cog_or_command = data.get("cc")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__ADD_RULE",
            "params": [str(guild), str(userid), allow_or_deny, str(who_or_what), cog_or_command],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/adddefaultrule", methods=["POST"])
def adddefaultrule(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["adddefaultrule"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["adddefaultrule"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=5)

    data = request.json

    userid = g.id
    allow_or_deny = data.get("ad")
    cog_or_command = data.get("cc")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__ADD_DEFAULT_RULE",
            "params": [str(guild), str(userid), allow_or_deny, cog_or_command],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/removerule", methods=["POST"])
def removerule(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["removerule"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["removerule"][session.get("id")] = datetime.datetime.now() + datetime.timedelta(
        seconds=5
    )

    data = request.json

    userid = g.id
    who_or_what = data.get("ww")
    cog_or_command = data.get("cc")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__REMOVE_RULE",
            "params": [str(guild), str(userid), str(who_or_what), cog_or_command],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/removedefaultrule", methods=["POST"])
def removedefaultrule(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["removedefaultrule"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["removedefaultrule"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=5)

    data = request.json

    userid = g.id
    cog_or_command = data.get("cc")

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_PERMISSIONS__REMOVE_DEFAULT_RULE",
            "params": [str(guild), str(userid), cog_or_command],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/api/<guild>/fetchaliases", methods=["GET"])
def fetchaliases(guild):
    if not session.get("id"):
        return jsonify({"status": 0, "message": "Not logged in"})

    if (
        end := app.cooldowns["fetchaliases"].get(
            session.get("id"), datetime.datetime.now() - datetime.timedelta(seconds=5)
        )
    ) > datetime.datetime.now():
        return jsonify(
            {
                "status": 0,
                "message": _("You are doing that too much.  Try again in {wait} seconds").format(
                    wait=(end - datetime.datetime.now()).seconds
                ),
            }
        )
    app.cooldowns["fetchaliases"][
        session.get("id")
    ] = datetime.datetime.now() + datetime.timedelta(seconds=5)

    userid = g.id

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_ALIASCC__FETCH_ALIASES",
            "params": [str(guild), str(userid)],
        }
        with app.lock:
            return get_result(app, requeststr)
    except:
        return jsonify({"status": 0, "message": "Not connected to bot"})


# --------------------------------------- API ---------------------------------------

# ----------------------------------- Third Party -----------------------------------


@blueprint.route("/api/webhook", methods=("POST",))
def webhook_route():
    # Webhook received
    if not request.is_json:
        # reject any requests that aren't json for now
        return jsonify(
            {"status": 0, "message": "Invalid formatting. This endpoint receives JSON only."}
        )
    payload = request.get_json()
    payload["origin"] = request.origin
    payload["headers"] = str(request.headers)
    # Pass header data here incase there was something
    # else the user needs for filtering
    payload["user_agent"] = str(request.user_agent)
    # User agent seems adequate enough for filtering
    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_WEBHOOKS__WEBHOOK_RECEIVE",
            "params": [payload],
        }
        with app.lock:
            return get_result(app, requeststr)
    except Exception:
        dashlog.warning("Error sending webhook info", exc_info=True)
        return jsonify({"status": 0, "message": "Not connected to bot"})


@blueprint.route("/third_party/spotify/callback")
def third_party_spotify_callback():
    # Spotify cog is owned by TrustyJAID

    try:
        code = request.args["code"]
        state = request.args["state"]
    except KeyError:
        return render_template(
            "third_party/spotify.html",
            context="3",
            msg="You must authenticate using a link given by bot.",
        )

    if not session.get("id"):
        session["login_redirect"] = {
            "route": f"api_blueprint.third_party_spotify_callback",
            "kwargs": {"code": code, "state": state},
        }
        return redirect(url_for("base_blueprint.login"))

    try:
        requeststr = {
            "jsonrpc": "2.0",
            "id": 0,
            "method": "DASHBOARDRPC_SPOTIFY__AUTHENTICATE_USER",
            "params": [str(g.id), code, state],
        }
        with app.lock:
            result = get_result(app, requeststr).json
        if result["status"] == 0:
            return render_template("third_party/spotify.html", context="2", msg=result["message"])
        if result["data"]["status"] == 0:
            return render_template(
                "third_party/spotify.html", context="3", msg=result["data"]["message"]
            )
        return render_template("third_party/spotify.html", context="1", msg="")
    except Exception as e:
        app.progress.print("".join(traceback.format_exception(type(e), e, e.__traceback__)))
        return render_template("third_party/spotify.html", context="2", msg=str(e))



import threading
import websocket


class Lock:
    def __init__(self):
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, *args):
        self.lock.release()


DEFAULTS = {
    "botname": "Red Discord Bot",
    "botavatar": "https://cdn.discordapp.com/icons/133049272517001216/a_aab012f3206eb514cac0432182e9e9ec.gif?size=1024",
    "botinfo": "Hello, welcome to the Red Discord Bot dashboard!  Here you can see basic information, commands list and even interact with your bot!  Unfortunately, this dashboard is not connected to any bot currently, so none of these features are available.  If you are the owner of the bot, please load the dashboard cog from Toxic Cogs.",
    "owner": "Cog Creators",
    "color": "red",
}

WS_URL = "ws://localhost:"
WS_EXCEPTIONS = (
    ConnectionRefusedError,
    websocket._exceptions.WebSocketConnectionClosedException,
    ConnectionResetError,
    ConnectionAbortedError,
)

ALLOWED_LOCALES = [
    "en",
    "af_ZA",
    "ar_SA",
    "bg_BG",
    "ca_ES",
    "cs_CZ",
    "da_DK",
    "de_DE",
    "el_GR",
    "es_ES",
    "fi_FI",
    "fr_FR",
    "he_IL",
    "hu_HU",
    "id_ID",
    "it_IT",
    "ja_JP",
    "ko_KR",
    "nl_NL",
    "no_NO",
    "pl_PL",
    "pt_BR",
    "pt_PT",
    "ro_RO",
    "ru_RU",
    "sk_SK",
    "sv_SE",
    "tr_TR",
    "uk_UA",
    "vi_VN",
    "zh_CN",
    "zh_HK",
    "zh_TW",
]



from reddash.app import app
from reddash.app.dashboard import blueprint
from flask import render_template, redirect, url_for, session, request, jsonify, Response, g
from flask_babel import _, refresh
from jinja2 import TemplateNotFound
import traceback
import websocket
import json
import time
import random
import logging
import datetime

dashlog = logging.getLogger("reddash")


@blueprint.route("/dashboard")
def dashboard():
    if not session.get("id"):
        return redirect(url_for("base_blueprint.login"))
    return render_template("dashboard.html")


@blueprint.route("/guild/<guild>")
def guild(guild):
    if not session.get("id"):
        return redirect(url_for("base_blueprint.login"))

    try:
        int(guild)
    except ValueError:
        raise ValueError("Guild ID must be integer")

    # We won't disconnect the websocket here, even if it fails, so that the main updating thread doesnt run into issues
    try:
        request = {
            "jsonrpc": "2.0",
            "id": random.randint(1, 1000),
            "method": "DASHBOARDRPC__GET_SERVER",
            "params": [int(g.id), int(guild)],
        }
        with app.lock:
            app.ws.send(json.dumps(request))
            result = json.loads(app.ws.recv())
            data = {}
            if "error" in result:
                if result["error"]["message"] == "Method not found":
                    data = {"status": 0, "message": "Not connected to bot"}
                else:
                    dashlog.error(result["error"])
                    data = {"status": 0, "message": "Something went wrong"}
            if isinstance(result["result"], dict) and result["result"].get("disconnected", False):
                data = {"status": 0, "message": "Not connected to bot"}
        if not data:
            data = {"status": 1, "data": result["result"]}
    except:
        data = {"status": 0, "message": "Not connected to bot"}
    return render_template("guild.html", data=data)



# -*- encoding: utf-8 -*-
"""
License: MIT
Copyright (c) 2019 - present AppSeed.us
"""

from flask import jsonify, render_template, redirect, request, url_for, session, g, make_response
from datetime import datetime, timedelta
from flask_babel import _, refresh

from reddash.app.base import blueprint
from reddash.app import app

import requests
import websocket
import json
import logging
import jwt

dashlog = logging.getLogger("reddash")


@blueprint.route("/error-<error>")
def route_errors(error):
    return render_template("errors/{}.html".format(error))


## Login & Registration


@blueprint.route("/callback", methods=["GET"])
def callback():
    try:
        code = request.args["code"]
    except KeyError:
        return render_template("login/login.html", status="1")
    requestobj = {
        "jsonrpc": "2.0",
        "id": 0,
        "method": "DASHBOARDRPC__GET_SECRET",
        "params": [],
    }
    if app.ws and app.ws.connected:
        with app.lock:
            try:
                app.ws.send(json.dumps(requestobj))
                result = json.loads(app.ws.recv())
            except (
                ConnectionRefusedError,
                websocket._exceptions.WebSocketConnectionClosedException,
                ConnectionResetError,
            ):
                return render_template("login/login.html", status="4")
            if "error" in result:
                if result["error"]["message"] == "Method not found":
                    return render_template("login/login.html", status="4")
                else:
                    dashlog.error(result["error"])
                    return render_template("login/login.html", status="5")
            if isinstance(result["result"], dict) and result["result"].get("disconnected", False):
                return render_template("login/login.html", status="4")
            secret = result["result"]["secret"]
    else:
        return render_template("login/login.html", status="4")
    redirectstr = app.variables["redirect"]
    data = {
        "client_id": int(app.variables["clientid"]),
        "client_secret": secret,
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": redirectstr,
        "scope": "identify",
    }
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    response = requests.post(
        "https://discordapp.com/api/v6/oauth2/token", data=data, headers=headers
    )
    try:
        token = response.json()["access_token"]
    except KeyError:
        dashlog.error(f"Failed to log someone in.\n{response.json()}")
        return render_template("login/login.html", status="2")
    new = requests.get(
        "https://discordapp.com/api/v6/users/@me", headers={"Authorization": f"Bearer {token}"},
    )
    new_data = new.json()
    if "id" in new_data:
        payload = {
            "userid": new_data["id"],
            "iat": datetime.utcnow(),
            "exp": datetime.utcnow() + timedelta(minutes=30),
        }
        token = jwt.encode(payload, app.jwt_secret_key, algorithm="HS256")
        session["id"] = token
        session[
            "avatar"
        ] = f"https://cdn.discordapp.com/avatars/{new_data['id']}/{new_data['avatar']}.png"
        session["username"] = new_data["username"]

        redirecting_to = "base_blueprint.index"
        arguments = {}
        if session.get("login_redirect"):
            redirecting_to = session["login_redirect"]["route"]
            arguments = session["login_redirect"]["kwargs"]
            del session["login_redirect"]

        return redirect(url_for(redirecting_to, **arguments))
    dashlog.error(f"Failed to obtain a user's profile.\n{new.json()}")
    return render_template("login/login.html", status="3")


@blueprint.route("/login", methods=["GET", "POST"])
def login():
    if not session.get("id"):
        return render_template("login/login.html", status="0")
    return redirect(url_for("base_blueprint.index"))


@blueprint.route("/logout")
def logout():
    del session["id"]
    return redirect(url_for("base_blueprint.login"))


@blueprint.route("/blacklisted")
def blacklisted():
    return render_template("errors/blacklisted.html")


@blueprint.route("/setcolor", methods=["POST"])
def set_color():
    resp = make_response(jsonify({"status": 1}))
    resp.set_cookie("color", request.json.get("color"))
    return resp


@blueprint.route("/index")
@blueprint.route("/")
def index():
    return render_template("index.html")


@blueprint.route("/commands")
def commands():
    data = app.commanddata
    prefix = app.variables.get("prefix", ["[p]"])

    return render_template(
        "pages/commands.html", cogs=[k["name"] for k in data], data=data, prefixes=prefix
    )


@blueprint.route("/credits")
def credits():
    return render_template("pages/credits.html")


## Errors


@blueprint.errorhandler(403)
def access_forbidden(error):
    return render_template("errors/403.html"), 403


@blueprint.errorhandler(404)
def not_found_error(error):
    return render_template("errors/404.html"), 404


@blueprint.errorhandler(500)
def internal_error(error):
    return render_template("errors/500.html"), 500



# -*- encoding: utf-8 -*-
"""
License: MIT
Copyright (c) 2019 - present AppSeed.us
"""

from flask import Blueprint

blueprint = Blueprint(
    "dashboard_blueprint", __name__, template_folder="templates", static_folder="static",
)



from importlib import import_module
from copy import copy
from os import path
import time
import json
import websocket
import jwt
import threading

from flask import render_template, redirect, request, url_for, session, g
from flask_babel import _
from rich import rule, columns, table as rtable, panel
from fuzzywuzzy import process

from reddash import __version__
from reddash.app.constants import DEFAULTS, WS_EXCEPTIONS


def register_blueprints(app):
    for module_name in ("base", "dashboard", "api"):
        module = import_module("reddash.app.{}.routes".format(module_name))
        app.register_blueprint(module.blueprint)

    @app.errorhandler(404)
    def page_not_found(e):
        return render_template("page-404.html"), 404

    @app.before_request
    def blockip():
        ip = request.environ.get("HTTP_X_FORWARDED_FOR") or request.remote_addr
        if request.path.startswith(("/static", "/api/stream", "/blacklisted")):
            return
        if not ip in app.blacklisted:
            g.id = get_user_id(app, request, session)
            if g.id is False and "id" in session:
                del session["id"]
        if ip in app.blacklisted:
            return redirect(url_for("base_blueprint.blacklisted"))


def apply_themes(app):
    """
    Add support for themes.

    If DEFAULT_THEME is set then all calls to
      url_for('static', filename='')
      will modfify the url to include the theme name

    The theme parameter can be set directly in url_for as well:
      ex. url_for('static', filename='', theme='')

    If the file cannot be found in the /static/<theme>/ location then
      the url will not be modified and the file is expected to be
      in the default /static/ location
    """

    @app.context_processor
    def override_url_for():
        return dict(url_for=_generate_url_for_theme)

    def _generate_url_for_theme(endpoint, **values):
        if endpoint.endswith("static"):
            themename = values.get("theme", None) or app.config.get("DEFAULT_THEME", None)
            if themename:
                theme_file = "{}/{}".format(themename, values.get("filename", ""))
                if path.isfile(path.join(app.static_folder, theme_file)):
                    values["filename"] = theme_file
            values["q"] = time.time()  # So Flask doesn't cache
        values["lang_code"] = session.get("lang_code", "en")
        return url_for(endpoint, **values)


def add_constants(app):
    @app.context_processor
    def inject_variables():
        variables = app.variables
        if not app.variables:
            variables = copy(DEFAULTS)
        variables["locales"] = app.config["LOCALE_DICT"]
        variables["safelocales"] = json.dumps(app.config["LOCALE_DICT"])
        variables["selectedlocale"] = session.get("lang_code")
        variables["color"] = request.cookies.get("color", variables["color"])
        variables = process_meta_tags(variables)
        return dict(version=__version__, **variables)


def initialize_babel(app, babel):
    @app.before_request
    def pull_locale():
        # Locale is determined in the following priority:
        #  lang_code argument
        #  lang_code session value
        #  default from browser
        args = request.args
        try:
            lang = args.get("lang_code", None)
        except AttributeError:
            lang = None

        if lang:
            # User had lang_code argument in request, lets check if its valid
            processed = process.extractOne(lang, app.config["LANGUAGES"])
            if processed[1] < 80:
                # Too low of a match, abort lang_code argument and go to session value
                lang = None
            else:
                # User had lang_code argument, and it closely matched a registered locale.  This will be used
                locale = processed[0]

        if not lang:
            # User either did not have lang_code argument or it wasnt able to match a registered locale.
            # Let's check if theres something in the session
            lang = session.get("lang_code")
            if lang:
                # User has a locale in session, that is already confirmed (otherwise would not be registered)
                locale = lang
            else:
                # User does not have lang_code argument, and does not have a locale stored in sesion.
                # Let's get the best one according to Flask
                locale = request.accept_languages.best_match(app.config["LANGUAGES"])

        # Locale variable is now the determined locale.  Let's save that so it will be used on next request as well
        session["lang_code"] = locale

    @babel.localeselector
    def get_locale():
        if not session.get("lang_code", None):
            session["lang_code"] = request.accept_languages.best_match(app.config["LANGUAGES"])
        return session["lang_code"]


def get_user_id(app, req, ses):
    try:
        payload = jwt.decode(ses["id"], app.jwt_secret_key, algorithms=["HS256"])
    except (jwt.ExpiredSignatureError, KeyError) as error:
        return False
    except (jwt.InvalidSignatureError, jwt.InvalidTokenError) as error:
        ip = req.environ.get("HTTP_X_FORWARDED_FOR") or req.remote_addr
        app.blacklisted.append(ip)
        thread = threading.Thread(target=notify_owner_of_blacklist, args=[app, ip])
        thread.start()
        return None
    return payload["userid"]


def process_meta_tags(variables):
    if "meta" not in variables:
        variables["meta"] = {"title": "", "icon": "", "description": "", "color": ""}

    if variables["meta"]["title"] == "":
        variables["meta"]["title"] = _("{name} Dashboard").format(name=variables["botname"])
    else:
        variables["meta"]["title"] = variables["meta"]["title"].replace(
            "{name}", variables["botname"]
        )

    if variables["meta"]["icon"] == "":
        variables["meta"]["icon"] = variables["botavatar"]

    if variables["meta"]["description"] == "":
        variables["meta"]["description"] = _(
            "Interactive dashboard to control and interact with {name}"
        ).format(name=variables["botname"])
    else:
        variables["meta"]["description"] = variables["meta"]["description"].replace(
            "{name}", variables["botname"]
        )

    if variables["meta"]["color"] == "":
        variables["meta"]["color"] = "#ff7575"

    return variables


def startup_message(app, progress, kwargs):
    table = rtable.Table(title="Settings")
    table.add_column("Setting", style="red", no_wrap=True)
    table.add_column("Value", style="blue", no_wrap=True)

    table.add_row("Version", __version__)
    table.add_row("Host", kwargs["host"])
    table.add_row("Webserver port", str(kwargs["port"]))
    table.add_row("RPC Port", str(app.rpcport))
    table.add_row("Update interval", str(app.interval))
    table.add_row("Environment", "Development" if kwargs["dev"] else "Production")
    table.add_row("Logging level", "Debug" if kwargs["debug"] else "Warning")

    progress.print(rule.Rule("Red Discord Bot Dashboard - Webserver"))
    disclaimer = "This is an instance of Red Discord Bot Dashboard,\ncreated by Neuro Assassin. This package is\nprotected under the MIT License. Any action\nthat will breach this license (including but not\nlimited to, removal of credits) may result in a\nDMCA takedown request, or other legal\nconsequences.\n\n\nYou can view the license at\nhttps://github.com/Cog-Creators/\nRed-Dashboard/blob/master/LICENSE."

    vartask = progress.add_task("Update variable task:", status="[bold blue]Starting[/bold blue]")
    cmdtask = progress.add_task("Update command task:", status="[bold blue]Starting[/bold blue]")
    vertask = progress.add_task("Update version task:", status="[bold blue]Starting[/bold blue]")
    contask = progress.add_task("RPC Connected", status="[bold blue]Starting[/bold blue]")

    progress.print(columns.Columns([panel.Panel(table), panel.Panel(disclaimer)], equal=True))
    return {"var": vartask, "cmd": cmdtask, "ver": vertask, "con": contask}


def initialize_websocket(app):
    app.ws = websocket.WebSocket()
    try:
        app.ws.connect(app.ws_url)
    except WS_EXCEPTIONS:
        app.ws.close()
        app.ws = None
        return False
    return True


def secure_send(app, request):
    try:
        app.ws.send(json.dumps(request))
    except WS_EXCEPTIONS:
        app.dashlog.warning("Connection reset")
        app.ws.close()
        app.ws = None
        return False

    try:
        result = json.loads(app.ws.recv())
    except WS_EXCEPTIONS:
        app.dashlog.warning("Connection reset")
        app.ws.close()
        app.ws = None
        return False
    else:
        return result


def check_for_disconnect(app, method, result):
    if "error" in result:
        if result["error"]["message"] == "Method not found":
            if method == "DASHBOARDRPC__GET_VARIABLES":
                app.variables = {}
        else:
            app.dashlog.error(result["error"])
        app.ws.close()
        app.ws = None
        return False
    if isinstance(result["result"], dict) and result["result"].get("disconnected", False):
        # Dashboard cog unloaded, disconnect
        if method == "DASHBOARDRPC__GET_VARIABLES":
            app.variables = {}
        app.ws.close()
        app.ws = None
        return False
    return True


def notify_owner_of_blacklist(app, ip):
    while True:
        if app.ws and app.ws.connected:
            request = {
                "jsonrpc": "2.0",
                "id": 0,
                "method": "DASHBOARDRPC__NOTIFY_OWNERS_OF_BLACKLIST",
                "params": [ip],
            }
            with app.lock:
                if not app.ws:
                    initialized = initialize_websocket(app)
                    if not initialized:
                        time.sleep(1)
                        continue

                result = secure_send(app, request)
                if not result or "error" in result:
                    time.sleep(1)
                    continue
                break
        time.sleep(1)



import time
import websocket
import threading

from reddash.app.utils import initialize_websocket, secure_send, check_for_disconnect


class TaskManager:
    def __init__(self, app, console, progress):
        self.threads = []
        self.app = app
        self.console = console
        self.progress = progress

    def update_variables(self, method, task):
        self.progress.update(task, status="[bold green]Running[/bold green]")
        self.progress.refresh()

        try:
            while True:
                # Different wait times based on method, commands should be called less due to how much data it is
                if method == "DASHBOARDRPC__GET_VARIABLES":
                    time.sleep(self.app.interval)
                else:
                    time.sleep(self.app.interval * 2)

                if not self.app.running:
                    self.progress.update(task, status="[bold red]Killed[/bold red]")
                    self.progress.refresh()
                    return

                request = {"jsonrpc": "2.0", "id": 0, "method": method, "params": []}
                with self.app.lock:
                    # This needs to be inside the lock, or both threads will create a websocket
                    if not self.app.ws:
                        initialized = initialize_websocket(self.app)
                        if not initialized:
                            continue

                    result = secure_send(self.app, request)
                    if not result:
                        continue

                    connected = check_for_disconnect(self.app, method, result)
                    if not connected:
                        continue

                if method == "DASHBOARDRPC__GET_VARIABLES":
                    self.app.variables = result["result"]
                    self.app.blacklisted = result["result"]["blacklisted"]
                else:
                    self.app.commanddata = result["result"]
                self.app.variables["disconnected"] = False
        except Exception as e:
            self.progress.update(task, status="[bold red]Stopped[/bold red]")
            self.progress.refresh()
            with self.console:
                if self.console.is_terminal:
                    self.console.print(self.progress._live_render.position_cursor())
                self.console.print_exception()
                if self.console.is_terminal:
                    self.console.print(self.progress._live_render)

    def update_version(self, task):
        self.progress.update(task, status="[bold green]Running[/bold green]")
        self.progress.refresh()

        try:
            while True:
                time.sleep(1)
                if not self.app.running:
                    self.progress.update(task, status="[bold red]Killed[/bold red]")
                    self.progress.refresh()
                    return
                if self.app.ws and self.app.ws.connected:
                    request = {
                        "jsonrpc": "2.0",
                        "id": 0,
                        "method": "DASHBOARDRPC__CHECK_VERSION",
                        "params": [],
                    }
                    with self.app.lock:
                        if not self.app.ws:
                            initialized = initialize_websocket(self.app)
                            if not initialized:
                                continue

                        result = secure_send(self.app, request)
                        if not result:
                            continue

                        if "error" in result:
                            continue

                        if "disconnected" in result["result"]:
                            continue

                        if (
                            result["result"]["v"] != self.app.rpcversion
                            and self.app.rpcversion != 0
                        ):
                            self.app.dashlog.warning(
                                "RPC websocket behind.  Closing and restarting..."
                            )
                            self.app.ws.close()
                            self.app.ws = websocket.WebSocket()
                            self.app.ws.connect(self.app.ws_url)
                        self.app.rpcversion = result["result"]["v"]
        except Exception as e:
            self.progress.update(task, status="[bold red]Stopped[/bold red]")
            self.progress.refresh()
            with self.console:
                if self.console.is_terminal:
                    self.console.print(self.progress._live_render.position_cursor())
                self.console.print_exception()
                if self.console.is_terminal:
                    self.console.print(self.progress._live_render)

    def check_if_connected(self, task):
        while True:
            if not self.app.running:
                self.app.ws.close()
                del self.app.ws
                self.progress.update(task, status="[bold red]Websocket Killed[/bold red]")
                self.progress.refresh()
                return
            time.sleep(0.1)
            if not (self.app.ws and self.app.ws.connected):
                self.progress.update(task, status="[bold red]Disconnected[/bold red]")
                self.progress.refresh()
            else:
                self.progress.update(task, status="[bold green]Connected[/bold green]")
                self.progress.refresh()

    def start_tasks(self, tasks):
        self.threads.append(
            threading.Thread(
                target=self.update_variables,
                args=["DASHBOARDRPC__GET_VARIABLES", tasks["var"]],
                daemon=True,
            )
        )
        self.threads.append(
            threading.Thread(
                target=self.update_variables,
                args=["DASHBOARDRPC__GET_COMMANDS", tasks["cmd"]],
                daemon=True,
            )
        )
        self.threads.append(
            threading.Thread(target=self.update_version, args=[tasks["ver"]], daemon=True)
        )
        self.threads.append(
            threading.Thread(target=self.check_if_connected, args=[tasks["con"]], daemon=True)
        )

        for t in self.threads:
            t.start()

    def stop_tasks(self):
        for t in self.threads:
            t.join()
